/*  Lziprecover - Data recovery program for lzip compressed files
    Copyright (C) 2008, 2009, 2010 Antonio Diaz Diaz.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
    Return values: 0 for a normal exit, 1 for environmental problems
    (file not found, invalid flags, I/O errors, etc), 2 to indicate a
    corrupt or invalid input file, 3 for an internal consistency error
    (eg, bug) which caused lzip to panic.
*/

#define _FILE_OFFSET_BITS 64

#include <cerrno>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <string>
#include <vector>
#include <fcntl.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/stat.h>

#include "arg_parser.h"
#include "lzip.h"
#include "decoder.h"


namespace {

const char * invocation_name = 0;
const char * const Program_name    = "Lziprecover";
const char * const program_name    = "lziprecover";
const char * const program_year    = "2010";

#ifdef O_BINARY
const int o_binary = O_BINARY;
#else
const int o_binary = 0;
#endif


void show_help() throw()
  {
  std::printf( "%s - Data recovery program for lzip compressed files.\n", Program_name );
  std::printf( "\nUsage: %s [options] file\n", invocation_name );
  std::printf( "Options:\n" );
  std::printf( "  -h, --help                 display this help and exit\n" );
  std::printf( "  -V, --version              output version information and exit\n" );
//  std::printf( "  -c, --create-recover-file  create a recover file\n" );
//  std::printf( "  -m, --merge                correct errors in file using a copy\n" );
  std::printf( "  -q, --quiet                suppress all messages\n" );
//  std::printf( "  -r, --recover              correct errors in file using a recover file\n" );
  std::printf( "  -R, --repair               try to repair a small error in file\n" );
  std::printf( "  -s, --split                split a multimember file in single-member files\n" );
//  std::printf( "  -u, --update               convert file from version 0 to version 1\n" );
  std::printf( "  -v, --verbose              be verbose (a 2nd -v gives more)\n" );
  std::printf( "\nReport bugs to lzip-bug@nongnu.org\n");
  std::printf( "Lzip home page: http://www.nongnu.org/lzip/lzip.html\n" );
  }


void show_version() throw()
  {
  std::printf( "%s %s\n", Program_name, PROGVERSION );
  std::printf( "Copyright (C) %s Antonio Diaz Diaz.\n", program_year );
  std::printf( "License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n" );
  std::printf( "This is free software: you are free to change and redistribute it.\n" );
  std::printf( "There is NO WARRANTY, to the extent permitted by law.\n" );
  }


int open_instream( const std::string & input_filename ) throw()
  {
  int infd = open( input_filename.c_str(), O_RDONLY | o_binary );
  if( infd < 0 )
    {
    if( verbosity >= 0 )
      std::fprintf( stderr, "%s: Can't open input file `%s': %s.\n",
                    program_name, input_filename.c_str(), std::strerror( errno ) );
    }
  else
    {
    struct stat in_stats;
    const int i = fstat( infd, &in_stats );
    if( i < 0 || !S_ISREG( in_stats.st_mode ) )
      {
      if( verbosity >= 0 )
        std::fprintf( stderr, "%s: input file `%s' is not a regular file.\n",
                      program_name, input_filename.c_str() );
      close( infd );
      infd = -1;
      }
    }
  return infd;
  }


int open_outstream( const std::string & output_filename ) throw()
  {
  int outfd = open( output_filename.c_str(),
                    O_CREAT | O_TRUNC | O_RDWR | o_binary,
                    S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH );
  if( outfd < 0 )
    {
    if( verbosity >= 0 )
      std::fprintf( stderr, "%s: Can't create output file `%s': %s.\n",
                    program_name, output_filename.c_str(), std::strerror( errno ) );
    }
  return outfd;
  }


bool next_filename( std::string & output_filename )
  {
  for( int i = 7; i >= 3; --i )			// "rec00001"
    {
    if( output_filename[i] < '9' ) { ++output_filename[i]; return true; }
    else output_filename[i] = '0';
    }
  return false;
  }


bool verify_header( const File_header & header )
  {
  if( !header.verify_magic() )
    {
    show_error( "bad magic number (file not in lzip format)." );
    return false;
    }
  if( header.version() == 0 )
    {
    show_error( "version 0 member format can't be recovered." );
    return false;
    }
  if( header.version() != 1 )
    {
    if( verbosity >= 0 )
      std::fprintf( stderr, "version %d member format not supported, newer %s needed.\n",
                    header.version(), program_name );
    return false;
    }
  return true;
  }


int do_split_file( const std::string & input_filename, uint8_t * & base_buffer )
  {
  const int hsize = File_header::size;
  const int tsize = File_trailer::size();
  const int buffer_size = 65536;
  const int base_buffer_size = tsize + buffer_size + hsize;
  base_buffer = new uint8_t[base_buffer_size];
  uint8_t * const buffer = base_buffer + tsize;

  const int infd = open_instream( input_filename );
  if( infd < 0 ) return 1;
  int size = readblock( infd, buffer, buffer_size + hsize ) - hsize;
  bool at_stream_end = ( size < buffer_size );
  if( size != buffer_size && errno )
    { show_error( "read error", errno ); return 1; }
  if( size <= tsize )
    { show_error( "file too short." ); return 2; }
  File_header header;
  for( int i = 0; i < File_header::size; ++i )
    header.data[i] = buffer[i];
  if( !verify_header( header ) ) return 2;

  std::string output_filename( "rec00001" ); output_filename += input_filename;
  int outfd = open_outstream( output_filename );
  if( outfd < 0 ) { close( infd ); return 1; }

  long long partial_member_size = 0;
  while( true )
    {
    int pos = 0;
    for( int newpos = 1; newpos <= size; ++newpos )
      if( buffer[newpos] == magic_string[0] &&
          buffer[newpos+1] == magic_string[1] &&
          buffer[newpos+2] == magic_string[2] &&
          buffer[newpos+3] == magic_string[3] )
        {
        long long member_size = 0;
        for( int i = 1; i <= 8; ++i )
          { member_size <<= 8; member_size += base_buffer[tsize+newpos-i]; }
        if( partial_member_size + newpos - pos == member_size )
          {						// header found
          const int wr = writeblock( outfd, buffer + pos, newpos - pos );
          if( wr != newpos - pos )
            { show_error( "write error", errno ); return 1; }
          if( close( outfd ) != 0 )
            { show_error( "error closing output file", errno ); return 1; }
          if( !next_filename( output_filename ) )
            { show_error( "too many members in file." ); close( infd ); return 1; }
          outfd = open_outstream( output_filename );
          if( outfd < 0 ) { close( infd ); return 1; }
          partial_member_size = 0;
          pos = newpos;
          }
        }

    if( at_stream_end )
      {
      const int wr = writeblock( outfd, buffer + pos, size + hsize - pos );
      if( wr != size + hsize - pos )
        { show_error( "write error", errno ); return 1; }
      break;
      }
    if( pos < buffer_size )
      {
      partial_member_size += buffer_size - pos;
      const int wr = writeblock( outfd, buffer + pos, buffer_size - pos );
      if( wr != buffer_size - pos )
        { show_error( "write error", errno ); return 1; }
      }
    std::memcpy( base_buffer, base_buffer + buffer_size, tsize + hsize );
    size = readblock( infd, buffer + hsize, buffer_size );
    at_stream_end = ( size < buffer_size );
    if( size != buffer_size && errno )
      { show_error( "read error", errno ); return 1; }
    }
  close( infd );
  if( close( outfd ) != 0 )
    { show_error( "error closing output file", errno ); return 1; }
  return 0;
  }


int split_file( const std::string & input_filename )
  {
  uint8_t * base_buffer;
  const int retval = do_split_file( input_filename, base_buffer );

  delete[] base_buffer;
  return retval;
  }


bool verify_single_member( const int fd, const long long file_size )
  {
  File_header header;
  if( lseek( fd, 0, SEEK_SET ) < 0 ||
      readblock( fd, header.data, File_header::size ) != File_header::size )
    { show_error( "error reading member header", errno ); return false; }
  if( !verify_header( header ) ) return false;

  File_trailer trailer;
  if( lseek( fd, -File_trailer::size(), SEEK_END ) < 0 ||
      readblock( fd, trailer.data, File_trailer::size() ) != File_trailer::size() )
    { show_error( "error reading member trailer", errno ); return false; }
  const long long member_size = trailer.member_size();
  if( member_size != file_size )
    {
    if( member_size < file_size &&
        lseek( fd, -member_size, SEEK_END ) > 0 &&
        readblock( fd, header.data, File_header::size ) == File_header::size &&
        verify_header( header ) )
      show_error( "input file has more than 1 member." );
    else
      show_error( "member size in input file trailer is corrupt." );
    return false;
    }
  return true;
  }


bool try_decompress( const int fd, const long long file_size,
                     long long * failure_pos = 0 )
  {
  std::vector< std::string > filenames;
  Pretty_print dummy( filenames );
  const int saved_verbosity = verbosity;
  verbosity = -1;

  try {
    Range_decoder rdec( fd );
    File_header header;
    rdec.reset_member_position();
    for( int i = 0; i < File_header::size; ++i )
      header.data[i] = rdec.get_byte();
    if( !rdec.finished() &&			// End Of File
        header.verify_magic() &&
        header.version() == 1 &&
        header.dictionary_size() >= min_dictionary_size &&
          header.dictionary_size() <= max_dictionary_size )
      {
      LZ_decoder decoder( header, rdec, -1 );

      if( decoder.decode_member( dummy ) == 0 &&
          rdec.member_position() == file_size )
        { verbosity = saved_verbosity; return true; }
      if( failure_pos ) *failure_pos = rdec.member_position();
      }
    }
  catch( std::bad_alloc )
    {
    show_error( "not enough memory. Find a machine with more memory." );
    std::exit( 1 );
    }
  catch( Error e ) {}
  verbosity = saved_verbosity;
  return false;
  }


bool copy_file( const int infd, const int outfd )
  {
  const int buffer_size = 65536;
  uint8_t * const buffer = new uint8_t[buffer_size];
  bool error = false;

  while( !error )
    {
    const int size = readblock( infd, buffer, buffer_size );
    if( size != buffer_size && errno )
      { show_error( "error reading input file", errno ); error = true; }
    if( size > 0 )
      {
      const int wr = writeblock( outfd, buffer, size );
      if( wr != size )
        { show_error( "error writing output file", errno ); error = true; }
      }
    if( size < buffer_size ) break;
    }
  delete[] buffer;
  return !error;
  }


std::string insert_fixed( std::string name ) throw()
  {
  if( name.size() > 4 && name.compare( name.size() - 4, 4, ".tlz" ) == 0 )
    name.insert( name.size() - 4, "-fixed" );
  else if( name.size() > 3 && name.compare( name.size() - 3, 3, ".lz" ) == 0 )
    name.insert( name.size() - 3, "-fixed" );
  else name += "-fixed.lz";
  return name;
  }


int repair_file( const std::string & input_filename )
  {
  const int infd = open_instream( input_filename );
  if( infd < 0 ) return 1;
  const long long isize = lseek( infd, 0, SEEK_END );
  if( isize < 0 )
    { show_error( "input file is not seekable", errno ); return 1; }
  if( isize < 36 )
    { show_error( "input file is too short." ); return 2; }
  if( !verify_single_member( infd, isize ) ) return 2;
  if( lseek( infd, 0, SEEK_SET ) < 0 )
    { show_error( "seek error in input file", errno ); return 1; }
  long long failure_pos = 0;
  if( try_decompress( infd, isize, &failure_pos ) )
    {
    if( verbosity >= 0 )
      std::printf( "Input file has no errors. No repairing is needed.\n" );
    return 0;
    }
  if( failure_pos >= isize - 8 ) failure_pos = isize - 8 - 1;
  if( failure_pos < File_header::size )
    { show_error( "can't repair error in input file." ); return 2; }

  const std::string output_filename( insert_fixed( input_filename ) );
  const int outfd = open_outstream( output_filename );
  if( outfd < 0 ) { close( infd ); return 1; }
  if( lseek( infd, 0, SEEK_SET ) < 0 )
    { show_error( "seek error in input file", errno ); return 1; }
  if( !copy_file( infd, outfd ) ) return 1;

  bool done = false;
  for( long long pos = failure_pos; pos >= File_header::size; --pos )
    {
    if( verbosity >= 0 )
      {
      std::printf( "Trying position %lld \r", pos );
      std::fflush( stdout );
      }
    uint8_t byte;
    if( lseek( outfd, pos, SEEK_SET ) < 0 ||
        readblock( outfd, &byte, 1 ) != 1 )
      { show_error( "error reading output file", errno ); return 1; }
    for( int i = 0; i < 255; ++i )
      {
      ++byte;
      if( lseek( outfd, pos, SEEK_SET ) < 0 ||
          writeblock( outfd, &byte, 1 ) != 1 ||
          lseek( outfd, 0, SEEK_SET ) < 0 )
        { show_error( "error writing output file", errno ); return 1; }
      if( try_decompress( outfd, isize ) )
        { done = true; break; }
      }
    if( done ) break;
    ++byte;
    if( lseek( outfd, pos, SEEK_SET ) < 0 ||
        writeblock( outfd, &byte, 1 ) != 1 )
      { show_error( "error writing output file", errno ); return 1; }
    }
  if( verbosity >= 0 ) std::printf( "\n" );

  if( close( outfd ) != 0 )
    { show_error( "error closing output file", errno ); return 1; }
  if( done )
    {
    if( verbosity >= 0 )
      std::printf( "Copy of input file repaired successfully.\n" );
    return 0;
    }
  else
    {
    std::remove( output_filename.c_str() );
    show_error( "too many errors in input file. Can't repair it." );
    return 2;
    }
  }

} // end namespace


void show_error( const char * const msg, const int errcode, const bool help ) throw()
  {
  if( verbosity >= 0 )
    {
    if( msg && msg[0] )
      {
      std::fprintf( stderr, "%s: %s", program_name, msg );
      if( errcode > 0 ) std::fprintf( stderr, ": %s", std::strerror( errcode ) );
      std::fprintf( stderr, "\n" );
      }
    if( help && invocation_name && invocation_name[0] )
      std::fprintf( stderr, "Try `%s --help' for more information.\n", invocation_name );
    }
  }


void internal_error( const char * const msg )
  {
  std::fprintf( stderr, "%s: internal error: %s.\n", program_name, msg );
  std::exit( 3 );
  }


int main( const int argc, const char * const argv[] )
  {
  enum Mode
    { m_none, m_create, m_merge, m_recover, m_repair, m_split, m_update };
  Mode program_mode = m_none;
  invocation_name = argv[0];

  const Arg_parser::Option options[] =
    {
    { 'h', "help",            Arg_parser::no  },
    { 'q', "quiet",           Arg_parser::no  },
    { 'R', "repair",          Arg_parser::no  },
    { 's', "split",           Arg_parser::no  },
    { 'v', "verbose",         Arg_parser::no  },
    { 'V', "version",         Arg_parser::no  },
    {  0 ,  0,                Arg_parser::no  } };

  Arg_parser parser( argc, argv, options );
  if( parser.error().size() )				// bad option
    { show_error( parser.error().c_str(), 0, true ); return 1; }

  int argind = 0;
  for( ; argind < parser.arguments(); ++argind )
    {
    const int code = parser.code( argind );
    if( !code ) break;					// no more options
    switch( code )
      {
      case 'h': show_help(); return 0;
      case 'q': verbosity = -1; break;
      case 'R': program_mode = m_repair; break;
      case 's': program_mode = m_split; break;
      case 'v': if( verbosity < 4 ) ++verbosity; break;
      case 'V': show_version(); return 0;
      default : internal_error( "uncaught option" );
      }
    }

  if( argind + 1 != parser.arguments() )
    { show_error( "you must specify exactly 1 file.", 0, true ); return 1; }

  if( program_mode == m_repair )
    return repair_file( parser.argument( argind ) );
  if( program_mode == m_split )
    return split_file( parser.argument( argind ) );
  show_error( "you must specify the operation to be performed on file.",
              0, true );
  return 1;
  }
